@@ -15,17 +15,15 @@
 # ******************************************************************************
 """Model quantization API"""
 
-import warnings
-import tensorflow as tf
-from keras.models import clone_model
-from keras.layers import Conv2D, SeparableConv2D, Dense, ReLU, Layer
+import tf_keras as keras
+from tf_keras.models import clone_model
+from tf_keras.layers import Conv2D, SeparableConv2D, Dense, ReLU, Layer
 
 from .quantization_ops import MaxQuantizer, MaxPerAxisQuantizer
 from .quantization_layers import (QuantizedConv2D, QuantizedSeparableConv2D,
                                   QuantizedDense, ActivationDiscreteRelu,
-                                  QuantizedActivation)
-from .utils import (invert_batchnorm_pooling, fold_batch_norms,
-                    _find_batchnorms_to_fold)
+                                  QuantizedActivation, QuantizedReLU)
+from .transforms import sequentialize, invert_batchnorm_pooling, fold_batchnorm
 from .cnn2snn_objects import cnn2snn_objects
 
 keras_neural_layers = (Conv2D, SeparableConv2D, Dense)
@@ -46,7 +44,7 @@ def quantize(model,
     This function returns a Keras model where the standard neural layers
     (Conv2D, SeparableConv2D, Dense) and the ReLU activations are replaced with
     CNN2SNN quantized layers (QuantizedConv2D, QuantizedSeparableConv2D,
-    QuantizedDense, ActivationDiscreteRelu).
+    QuantizedDense, QuantizedRelu).
 
     Several transformations are applied to the model:
     - the order of MaxPool and BatchNormalization layers are inverted so that
@@ -57,7 +55,7 @@ def quantize(model,
     retrained for a few epochs to recover any accuracy loss.
 
     Args:
-        model (tf.keras.Model): a standard Keras model
+        model (keras.Model): a standard Keras model
         weight_quantization (int): sets all weights in the model to have
             a particular quantization bitwidth except for the weights in the
             first layer.
@@ -82,22 +80,20 @@ def quantize(model,
             returns the quantizer to use.
 
     Returns:
-        tf.keras.Model: a quantized Keras model
+        keras.Model: a quantized Keras model
     """
 
     # Overrides input weight quantization if None
     if input_weight_quantization is None:
         input_weight_quantization = weight_quantization
 
-    # Apply Model transformations, obtaining strictly equivalent models at
-    # inference time.
-    # Invert batch normalization and pooling
-    model_t = invert_batchnorm_pooling(model)
-    # Raise warning
-    _check_inconsistent_fold_BN(model_t, fold_BN)
     if fold_BN:
+        # Identify sequential branches
+        seq_model = sequentialize(model)
+        # Invert batch normalization and pooling
+        model_t = invert_batchnorm_pooling(seq_model)
         # Fold batch norm layers with corresponding neural layers
-        model_t = fold_batch_norms(model_t)
+        model_t = fold_batchnorm(model_t)
     else:
         # Keep original model without transformation
         model_t = model
@@ -119,7 +115,7 @@ def quantize(model,
                                                quantizer_function)
         return layer.__class__.from_config(layer.get_config())
 
-    with tf.keras.utils.custom_object_scope(cnn2snn_objects):
+    with keras.utils.custom_object_scope(cnn2snn_objects):
         new_model = clone_model(model_t, clone_function=replace_layer)
     new_model.set_weights(model_t.get_weights())
 
@@ -140,9 +136,9 @@ def quantize_layer(model, target_layer, bitwidth, quantizer_function=None):
     Examples:
 
         >>> # Quantize a layer of a native Keras model
-        >>> model = tf.keras.Sequential([
-        ...     tf.keras.layers.Dense(5, input_shape=(3,)),
-        ...     tf.keras.layers.Softmax()])
+        >>> model = keras.Sequential([
+        ...     keras.layers.Dense(5, input_shape=(3,)),
+        ...     keras.layers.Softmax()])
         >>> model_quantized = cnn2snn.quantize_layer(model,
         ...                                          target_layer=0,
         ...                                          bitwidth=4)
@@ -157,7 +153,7 @@ def quantize_layer(model, target_layer, bitwidth, quantizer_function=None):
         2
 
     Args:
-        model (tf.keras.Model): a standard Keras model
+        model (keras.Model): a standard Keras model
         target_layer: a standard or quantized Keras layer to be
             converted, or the index or name of the target layer.
         bitwidth (int): the desired quantization bitwidth. If zero, no
@@ -167,7 +163,7 @@ def quantize_layer(model, target_layer, bitwidth, quantizer_function=None):
             returns the quantizer to use.
 
     Returns:
-        tf.keras.Model: a quantized Keras model
+        keras.Model: a quantized Keras model
 
     Raises:
         ValueError: In case of invalid target layer
@@ -197,7 +193,7 @@ def quantize_layer(model, target_layer, bitwidth, quantizer_function=None):
             return layer.__class__.from_config(layer.get_config())
         return layer.__class__.from_config(layer.get_config())
 
-    with tf.keras.utils.custom_object_scope(cnn2snn_objects):
+    with keras.utils.custom_object_scope(cnn2snn_objects):
         new_model = clone_model(model, clone_function=replace_layer)
     new_model.set_weights(model.get_weights())
 
@@ -213,7 +209,7 @@ def _convert_to_quantized_layer(layer, bitwidth, quantizer_function=None):
     MaxPerAxisQuantizer if quantizer_function is not specified.
 
     Args:
-        layer (tf.keras.Layer): a standard Keras (Conv2D, SeparableConv2D or
+        layer (keras.Layer): a standard Keras (Conv2D, SeparableConv2D or
             Dense) or quantized (QuantizedConv2D, QuantizedSeparableConv2D,
             QuantizedDense) layer.
         bitwidth (int): the desired weight quantization bitwidth. If zero, the
@@ -239,7 +235,7 @@ def _convert_to_quantized_layer(layer, bitwidth, quantizer_function=None):
         return layer.__class__.from_config(config)
 
     # Set quantizer with expected bitwidth
-    if not 'quantizer' in config:
+    if 'quantizer' not in config:
         quantizer = MaxPerAxisQuantizer(bitwidth=bitwidth)
         if quantizer_function is not None:
             quantizer = quantizer_function(layer, quantizer)
@@ -263,7 +259,7 @@ def _convert_to_quantized_layer(layer, bitwidth, quantizer_function=None):
         return QuantizedConv2D.from_config(config)
 
     if isinstance(layer, SeparableConv2D):
-        if not 'quantizer_dw' in config:
+        if 'quantizer_dw' not in config:
             quantizer = MaxQuantizer(bitwidth=bitwidth)
             if quantizer_function is not None:
                 quantizer = quantizer_function(layer, quantizer)
@@ -286,7 +282,7 @@ def _quantize_activation_layer(layer, bitwidth):
     layer.
 
     Args:
-        layer (tf.keras.Layer): an activation layer (ReLU or CNN2SNN quantized
+        layer (keras.Layer): an activation layer (ReLU or CNN2SNN quantized
             activation)
         bitwidth (int): the desired quantization bitwidth. If zero, the ReLU
             layer will be returned as it is.
@@ -307,47 +303,18 @@ def _quantize_activation_layer(layer, bitwidth):
         return layer.__class__.from_config(layer.get_config())
 
     if isinstance(layer, ReLU):
-        return ActivationDiscreteRelu(bitwidth, name=layer.name)
+        if layer.threshold != 0:
+            print(f"Skipping ReLU layer {layer.name} with non-zero threshold")
+            return layer.__class__.from_config(layer.get_config())
+        # Instantiate a QuantizedReLU using the original ReLU max_value
+        return QuantizedReLU(bitwidth, layer.max_value, name=layer.name)
+
+    if isinstance(layer, ActivationDiscreteRelu):
+        # Evaluate the ActivationDiscreteRelu maximum value
+        max_value = layer.step * layer.levels
+        # Convert it to a QuantizedReLU
+        return QuantizedReLU(bitwidth, max_value, name=layer.name)
 
     config = layer.get_config()
     config['bitwidth'] = bitwidth
     return layer.__class__.from_config(config)
-
-
-def _check_inconsistent_fold_BN(model_inverted_mp_bn, fold_BN):
-    """Raises a warning message if 'fold_BN' argument is not consistent with the
-    model.
-
-    Warning is raised in two configurations:
-      - a quantized neural layer with a BN layer is present and fold_BN = True.
-      - a float neural layer with a BN layer is present and fold_BN = False.
-
-    In both cases, we advise to retrain on a few batches to align BN parameters
-    with the quantized weights.
-
-    Args:
-        model_inverted_mp_bn (:obj:`tf.keras.Model`): a Keras model to check.
-            This model must have BN layers directly after neural layers. This is
-            done by first applying the "invert_batchnorm_pooling" transformation
-        fold_BN (bool): whether BN layers must be folded or not
-    """
-    warnings.simplefilter('always', UserWarning)
-
-    neural_layers_to_fold_bn, _ = _find_batchnorms_to_fold(model_inverted_mp_bn)
-    for neural_layer in neural_layers_to_fold_bn:
-        if fold_BN and type(neural_layer) in cnn2snn_neural_layers:
-            warnings.warn(
-                "The model contains the quantized neural layer (layer "
-                f"{neural_layer.name} of type {neural_layer.__class__.__name__}"
-                f") followed by a BatchNormalization layer. This configuration"
-                " is not consistent with fold_BN=True. This might require "
-                "retraining on a few batches to align BatchNormalization "
-                "parameters with the quantized weights.")
-        elif not fold_BN and type(neural_layer) in keras_neural_layers:
-            warnings.warn(
-                "The model contains the float neural layer (layer "
-                f"{neural_layer.name} of type {neural_layer.__class__.__name__}"
-                f") followed by a BatchNormalization layer. This configuration"
-                " is not consistent with fold_BN=False. This might require "
-                "retraining on a few batches to align BatchNormalization "
-                "parameters with the new quantized weights.")
