@@ -18,11 +18,11 @@
 
 # Tensorflow imports
 import tensorflow as tf
-from keras import backend as K
-from keras import layers
-from keras.utils.generic_utils import serialize_keras_object
-from keras.utils import conv_utils  # pylint: disable=no-name-in-module
-from tensorflow.python.ops import nn  # pylint: disable=no-name-in-module
+from tf_keras import backend as K
+from tf_keras import layers
+from tf_keras.utils import serialize_keras_object
+from tf_keras.src.utils import conv_utils
+from tensorflow.python.ops import nn
 from .quantization_ops import get as get_quantizer
 from .quantization_ops import ceil_through
 
@@ -30,8 +30,7 @@ from .quantization_ops import ceil_through
 def _check_unsupported_args(kwargs, unsupported_args):
     """Raises error if unsupported argument are present in kwargs.
 
-    For now, 4 arguments are unsupported: 'data_format', 'activation',
-        'dilation_rate', 'depth_mutiplier'.
+    Unsupported arguments: 'data_format', 'activation', 'depth_mutiplier'.
 
     Args:
         kwargs (dictionary): keyword arguments to check.
@@ -55,29 +54,28 @@ class QuantizedConv2D(layers.Conv2D):
         quantizer (:obj:`cnn2snn.WeightQuantizer`): the quantizer
             to apply during the forward pass.
         strides (integer, or tuple of integers, optional): strides of the
-            convolution along height and width.
+            convolution along spatial dimensions.
         padding (str, optional): one of 'valid' or 'same'.
         use_bias (boolean, optional): whether the layer uses a bias vector.
-        kernel_initializer (str, or a :obj:`tf.keras.initializer`, optional):
+        kernel_initializer (str, or a :obj:`keras.initializer`, optional):
             initializer for the weights matrix.
-        bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
+        bias_initializer (str, or a :obj:`keras.initializer`, optional):
             initializer for the bias vector.
-        kernel_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        kernel_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the weights.
-        bias_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        bias_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the bias.
-        activity_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        activity_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the output of the layer.
-        kernel_constraint (str, or a :obj:`tf.keras.constraint`, optional):
+        kernel_constraint (str, or a :obj:`keras.constraint`, optional):
             constraint applied to the weights.
-        bias_constraint (str, or a :obj:`tf.keras.constraint`, optional):
+        bias_constraint (str, or a :obj:`keras.constraint`, optional):
             constraint applied to the bias.
 
     """
     unsupported_args = {
         'data_format': 'channels_last',
         'activation': 'linear',
-        'dilation_rate': (1, 1),
         'groups': 1
     }
 
@@ -128,8 +126,8 @@ class QuantizedConv2D(layers.Conv2D):
         """
         input_shape = inputs.shape
 
-        outputs = self._convolution_op(inputs,
-                                       self.quantizer.quantize(self.kernel))
+        outputs = self.convolution_op(inputs,
+                                      self.quantizer.quantize(self.kernel))
         if self.use_bias:
             output_rank = outputs.shape.rank
             # Handle multiple batch dimensions.
@@ -140,10 +138,8 @@ class QuantizedConv2D(layers.Conv2D):
                                           self.bias,
                                           data_format=self._tf_data_format)
 
-                outputs = conv_utils.squeeze_batch_dims(outputs,
-                                                        _apply_fn,
-                                                        inner_rank=self.rank +
-                                                        1)
+                outputs = conv_utils.squeeze_batch_dims(
+                    outputs, _apply_fn, inner_rank=self.rank + 1)
             else:
                 outputs = tf.nn.bias_add(outputs,
                                          self.bias,
@@ -173,24 +169,24 @@ class QuantizedDepthwiseConv2D(layers.DepthwiseConv2D):
     Args:
         kernel_size (a tuple of integer): the kernel spatial dimensions.
         strides (integer, or tuple of integers, optional): strides of the
-            convolution along height and width.
+            convolution along spatial dimensions.
         quantizer (:obj:`cnn2snn.WeightQuantizer`): the quantizer
             to apply during the forward pass.
         padding (str, optional): One of 'valid' or 'same'.
         use_bias (boolean, optional): whether the layer uses a bias vector.
-        depthwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
+        depthwise_initializer (str, or a :obj:`keras.initializer`, optional):
             initializer for the weights matrix.
-        bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
+        bias_initializer (str, or a :obj:`keras.initializer`, optional):
             initializer for the bias vector.
-        depthwise_regularizer (str, or a :obj:`tf.keras.initializer`, optional):
+        depthwise_regularizer (str, or a :obj:`keras.initializer`, optional):
             regularization applied to the weights.
-        bias_regularizer (str, or a :obj:`tf.keras.initializer`, optional):
+        bias_regularizer (str, or a :obj:`keras.initializer`, optional):
             regularization applied to the bias.
-        activity_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        activity_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the output of the layer.
-        depthwise_constraint (str, or a :obj:`tf.keras.initializer`, optional):
+        depthwise_constraint (str, or a :obj:`keras.initializer`, optional):
             constraint applied to the weights.
-        bias_constraint (str, or a :obj:`tf.keras.initializer`, optional):
+        bias_constraint (str, or a :obj:`keras.initializer`, optional):
             constraint applied to the bias.
 
     """
@@ -271,19 +267,19 @@ class QuantizedDense(layers.Dense):
         use_bias (boolean, optional): whether the layer uses a bias vector.
         quantizer (:obj:`cnn2snn.WeightQuantizer`): the quantizer
             to apply during the forward pass.
-        kernel_initializer (str, or a :obj:`tf.keras.initializer`, optional):
+        kernel_initializer (str, or a :obj:`keras.initializer`, optional):
             initializer for the weights matrix.
-        bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
+        bias_initializer (str, or a :obj:`keras.initializer`, optional):
             initializer for the bias vector.
-        kernel_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        kernel_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the weights.
-        bias_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        bias_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the bias.
-        activity_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        activity_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the output of the layer.
-        kernel_constraint (str, or a :obj:`tf.keras.constraint`, optional):
+        kernel_constraint (str, or a :obj:`keras.constraint`, optional):
             constraint applied to the weights.
-        bias_constraint (str, or a :obj:`tf.keras.constraint`, optional):
+        bias_constraint (str, or a :obj:`keras.constraint`, optional):
             constraint applied to the bias.
 
     """
@@ -329,7 +325,7 @@ class QuantizedDense(layers.Dense):
 
         """
         if inputs.dtype.base_dtype != self._compute_dtype_object.base_dtype:
-            inputs = tf.cast(inputs, dtype=self._compute_dtype_object)  # pylint: disable=unexpected-keyword-arg, no-value-for-parameter
+            inputs = tf.cast(inputs, dtype=self._compute_dtype_object)
 
         if isinstance(inputs, tf.RaggedTensor):
             raise TypeError(f"The inputs of a Dense Layer must be a uniform \
@@ -386,35 +382,34 @@ class QuantizedSeparableConv2D(layers.SeparableConv2D):
         quantizer_dw (:obj:`cnn2snn.WeightQuantizer`, optional): the
             depthwise quantizer to apply during the forward pass.
         strides (integer, or tuple of integers, optional): strides of the
-            convolution along height and width.
+            convolution along spatial dimensions.
         padding (str, optional): One of 'valid' or 'same'.
         use_bias (boolean, optional): Whether the layer uses a bias vector.
-        depthwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
+        depthwise_initializer (str, or a :obj:`keras.initializer`, optional):
             initializer for the depthwise kernel.
-        pointwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
+        pointwise_initializer (str, or a :obj:`keras.initializer`, optional):
             initializer for the pointwise kernel.
-        bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
+        bias_initializer (str, or a :obj:`keras.initializer`, optional):
             initializer for the bias vector.
-        depthwise_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        depthwise_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the depthwise kernel.
-        pointwise_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        pointwise_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the pointwise kernel.
-        bias_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        bias_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the bias.
-        activity_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
+        activity_regularizer (str, or a :obj:`keras.regularizer`, optional):
             regularization applied to the output of the layer.
-        depthwise_constraint (str, or a :obj:`tf.keras.constraint`, optional):
+        depthwise_constraint (str, or a :obj:`keras.constraint`, optional):
             constraint applied to the depthwise kernel.
-        pointwise_constraint (str, or a :obj:`tf.keras.constraint`, optional):
+        pointwise_constraint (str, or a :obj:`keras.constraint`, optional):
             constraint applied to the pointwise kernel.
-        bias_constraint (str, or a :obj:`tf.keras.constraint`, optional):
+        bias_constraint (str, or a :obj:`keras.constraint`, optional):
             constraint applied to the bias.
 
     """
     unsupported_args = {
         'data_format': 'channels_last',
         'activation': 'linear',
-        'dilation_rate': (1, 1),
         'depth_multiplier': 1
     }
 
@@ -437,7 +432,6 @@ class QuantizedSeparableConv2D(layers.SeparableConv2D):
                  pointwise_constraint=None,
                  bias_constraint=None,
                  **kwargs):
-        # pylint: disable=too-many-locals
 
         _check_unsupported_args(kwargs, self.unsupported_args)
         self.quantizer = get_quantizer(quantizer)
@@ -510,14 +504,9 @@ class QuantizedSeparableConv2D(layers.SeparableConv2D):
 class QuantizedActivation(layers.Layer):
     """Base class for quantized activation layers.
 
-    This base class must be overloaded as well as the three @property
-    functions:
+    This base class must be overloaded as well as the `step` @property function.
 
-    - `threshold`
-    - `step_height`
-    - `step_width`
-
-    These @property functions must return TensorFlow objects (e.g. tf.Tensor
+    This @property function must return a TensorFlow object (e.g. tf.Tensor
     or tf.Variable) of scalar values. The `.numpy()` method must be callable on
     them. They can be fixed at initialization or can be trainable variables.
 
@@ -546,30 +535,32 @@ class QuantizedActivation(layers.Layer):
     def quantized_activation(self, x):
         """Evaluates the quantized activations for the specified input Tensor.
 
-        The quantization is defined by three parameters:
+        Activations will be clipped to a quantization range, and quantized to a
+        number of values defined by the bitwidth: N = (2^bitwidth - 1) values
+        plus zero.
 
-        - the activation threshold, 'threshold'
-        - the quantization step size, 'step_height'
-        - the step width, 'step_width'
+        The quantization is defined by a single step parameter, that defines
+        the interval between two quantized values.
+
+        A quantization threshold set to half the quantization step is used to
+        evaluate the quantization intervals, to make sure that each quantized
+        value is exactly in the middle of its quantization interval, thus
+        minimizing the quantization error.
 
         For any potential x, the activation output is as follows:
 
         - if x <= threshold, activation is zero
-        - if threshold < x <= threshold + step_width, activation is
-          step_height
-        - if threshold + step_width < x <= threshold + 2*step_width,
-          activation is 2*step_height
-        - and so on...
-        - if x > threshold + levels*step_width, activation is
-          levels*step_height
+        - if threshold + (n - 1) * step < x <= threshold + n * step,
+          activation is n * step
+        - if x > threshold + levels * step, activation is levels * step
 
         Args:
             x (:obj:`tensorflow.Tensor`): the input values.
 
         """
-        act = ceil_through((x - self.threshold) / self.step_width)
+        act = ceil_through((x - self.threshold) / self.step)
         clip_act = tf.clip_by_value(act, 0, self.levels)
-        return self.step_height * clip_act
+        return self.step * clip_act
 
     def call(self, inputs, *args, **kwargs):
         """Evaluates the quantized activations for the specified input Tensor.
@@ -583,17 +574,14 @@ class QuantizedActivation(layers.Layer):
 
     @property
     def threshold(self):
-        """Returns the activation threshold"""
-        raise NotImplementedError()
+        """The quantization threshold is equal to half the quantization step to
+        better approximate the ReLU.
+        """
+        return self.step * 0.5
 
     @property
-    def step_height(self):
-        """Returns the step height of the quantized activation"""
-        raise NotImplementedError()
-
-    @property
-    def step_width(self):
-        """Returns the step width of the quantized activation"""
+    def step(self):
+        """Returns the interval between two quantized activation values"""
         raise NotImplementedError()
 
     def get_config(self):
@@ -605,8 +593,13 @@ class QuantizedActivation(layers.Layer):
 class ActivationDiscreteRelu(QuantizedActivation):
     """A discrete ReLU Keras Activation.
 
-    Activations will be quantized and will have 2^bitwidth values in the range
-    [0,6].
+    For bitwidth 1 or 2:
+
+        - threshold is 0.5 and step is 1
+
+    For bithwidth > 2, with N = 2^bitwidth - 1:
+
+        - threshold is 3 / N and step is 6 / N
 
     Args:
         bitwidth (int): the activation bitwidth.
@@ -617,79 +610,42 @@ class ActivationDiscreteRelu(QuantizedActivation):
         super().__init__(bitwidth, **kwargs)
 
         relumax = min(self.levels, 6)
-        self.step_height_ = tf.constant(relumax / self.levels)
+        self.step_ = tf.constant(relumax / self.levels)
 
     @property
-    def threshold(self):
-        return 0.5 * self.step_height_
-
-    @property
-    def step_height(self):
-        return self.step_height_
-
-    @property
-    def step_width(self):
-        return self.step_height_
+    def step(self):
+        return self.step_
 
 
 class QuantizedReLU(QuantizedActivation):
-    """A Trainable Quantized ReLU Keras Activation.
+    """A configurable Quantized ReLU Keras Activation.
 
-    Activations will be clipped to a trainable range, and quantized to a number
-    of values defined by the bitwidth: N = (2^bitwidth - 1) values plus zero
-
-    More specifically, this class uses two trainable variables:
-
-    - threshold represents the lower bound of the activation range,
-    - step_height represents the step between two quantized activation values.
-
-    The activation range is therefore [threshold, tN_k], with:
-
-        tN_k = threshold + N * step_height = (2^bitwidth - 1) * step_height
-
-    In other words:
-
-    - inputs below threshold will result in no activation
-    - inputs between threshold and threshold + tN_k will be ceiled to the nearest
-      threshold + n * step_height, and result in a activation of n * step_height
-    - inputs above threshold + tN_k will result in a activation of N * step_height
+    In addition to the quantization bitwidth, this class can be initialized
+    with a max_value parameter corresponding to the ReLU maximum value.
 
     Args:
         bitwidth (int): the activation bitwidth.
-
+        max_value (float): the initial max_value
     """
 
-    def __init__(self, bitwidth=1, **kwargs):
+    def __init__(self, bitwidth=1, max_value=None, **kwargs):
         super().__init__(bitwidth, **kwargs)
 
-        # To be compatible with models trained with ActivationDiscreteRelu, we
-        # keep the same rule for the initial value of the activation range
-        # relumax and we deduce the initial values of the threshold and
-        # step_height from it
-        relumax = min(self.levels, 6)
-        step_height_initializer = relumax / self.levels
-        # Create lower bound variable and initialize it
-        threshold_initializer = step_height_initializer / 2.
-        self.threshold_ = tf.Variable(name=f"{self.name}/threshold",
-                                      initial_value=threshold_initializer,
-                                      dtype=tf.float32,
-                                      trainable=True)
-        # We will actually train the rescaled step, whose value is more
-        # consistent between different bitwidth values
-        step_initializer = (2.**self.bitwidth) * step_height_initializer / 16
-        self.step_ = tf.Variable(name=f"{self.name}/step",
-                                 initial_value=step_initializer,
-                                 dtype=tf.float32,
-                                 trainable=True)
+        # The max_value defaults to None for ReLU, which means unclipped.
+        # For QuantizedReLU we need to set a value, by default we align
+        # on the bitwidth.
+        if max_value is None:
+            max_value = 2.**bitwidth - 1
+        self.max_value_ = tf.constant(max_value, dtype=tf.float32)
+        # Evaluate the quantization step
+        self.step_ = tf.constant(self.max_value_ / self.levels,
+                                 dtype=tf.float32)
 
     @property
-    def step_height(self):
-        return self.step_width
+    def step(self):
+        return self.step_
 
-    @property
-    def threshold(self):
-        return self.threshold_
-
-    @property
-    def step_width(self):
-        return 16 * self.step_ / 2.**self.bitwidth
+    def get_config(self):
+        config = super().get_config()
+        config.update({'max_value': self.max_value_.numpy()})
+        return config
