@@ -20,7 +20,7 @@
 from collections import namedtuple
 import numpy as np
 import tensorflow as tf
-import keras.layers as layers
+import tf_keras.layers as layers
 from akida import LayerType
 
 KerasLayerParams = namedtuple(
@@ -48,7 +48,7 @@ def set_layer_variables(layer_ak,
     # Convert weights
     weights, weights_scale_factor = \
         convert_layer_weights(layer_ak.parameters.layer_type, layer_neural,
-                                layer_ak.input_dims)
+                              layer_ak.input_dims)
 
     # Convert fire thresholds and steps
     th_fire, th_step = _convert_layer_thresholds(layer_ak.parameters.layer_type,
@@ -77,7 +77,7 @@ def convert_layer_weights(layer_type, layer_k, input_dims=None):
 
     Args:
         layer_type (str): type of the layer Akida.
-        layer_k (:obj:`tf.keras.Layer`): the Keras layer from which the weights
+        layer_k (:obj:`keras.Layer`): the Keras layer from which the weights
             will be converted.
         input_dims (tuple): input dimensions of the Keras layer. This argument
             is only required for a fullyConnected Akida layer.
@@ -104,9 +104,7 @@ def _convert_conv_weights(layer, input_conv=False):
     weights = layer.get_weights()[0]
     wq, delta = quantize_weights(layer.quantizer, weights)
 
-    # Transpose weights to get from Keras HWCN to Akida WHCN
-    # and multiply by the weights scale factor.
-    wq_akida = wq.transpose((1, 0, 2, 3)) * delta
+    wq_akida = wq * delta
     wq_akida = np.round(wq_akida).astype(np.int8)
 
     # Flip W and H dimensions for conv. kernels (not input conv.)
@@ -120,7 +118,7 @@ def _convert_separable_conv_weights(layer):
     # Quantize depthwise weights
     weights = layer.get_weights()[0]
     wq_dw, delta_dw = quantize_weights(layer.quantizer_dw, weights)
-    w_ak_dw = wq_dw.transpose((1, 0, 2, 3)) * delta_dw
+    w_ak_dw = wq_dw * delta_dw
     w_ak_dw = np.round(w_ak_dw).astype(np.int8)
     w_ak_dw = np.flip(w_ak_dw, axis=[0, 1])
 
@@ -144,21 +142,18 @@ def _convert_dense_weights(layer, input_dims):
     wq, delta = quantize_weights(layer.quantizer, weights)
 
     # retrieve input dimensions from Akida's layer
-    inwidth, inheight, inchans = input_dims
-    # Kernels in the fully connected are in the (HxWxC,N) format, more
-    # specifically in each neuron data is laid out in the H,W,C format. In
-    # Akida we expect a kernel in the (N,CxHxW,1,1), where the data in each
-    # neuron is laid out in the W,H,C format as the input dimensions are
-    # set.
-    # So the operations done in order to obtain the akida fully connected
-    # kernel are:
-    # 1. reshape to H,W,C,N to split data across dimensions
-    # 2. transpose dimensions to obtain C,H,W,N
-    # 3. reshape to: 1,1,CxHxW,N
+    inX, inY, inchans = input_dims
+    # When a fully connected layer follows a convolutional layer, we need to modify
+    # the way the weights are laid out because the Keras and Akida flatten operations
+    # that happen on spatial dimensions are inverted (row-major versus col-major).
+    # We therefore need to:
+    # - reshape the Keras (FxN) weights to match the input shape (XxYxCxN),
+    # - transpose to the equivalent akida shape (CxYxXxN),
+    # - reshape back to a flat shape (1x1xFxN).
     #
-    wq_akida = wq.reshape(inheight, inwidth, inchans, layer.units) \
-        .transpose(2, 0, 1, 3) \
-        .reshape(1, 1, inchans * inheight * inwidth, layer.units)
+    wq_akida = wq.reshape(inX, inY, inchans, layer.units) \
+        .transpose(2, 1, 0, 3) \
+        .reshape(1, 1, inchans * inY * inX, layer.units)
     # Multiply by delta, round and cast to int
     wq_akida = wq_akida * delta
     wq_akida = np.round(wq_akida).astype(np.int8)
@@ -179,9 +174,9 @@ def _prepare_threshold_calculations(layer_ak_type,
 
     Args:
         layer_ak_type (:obj:`akida.LayerType`): the type of the Akida layer.
-        layer_neural (:obj:`tf.keras.Layer`): the Keras neural layer.
-        layer_activation (:obj:`tf.keras.Layer`): the Keras activation layer.
-        layer_globalavgpool (:obj:`tf.keras.Layer`): the optional Keras global
+        layer_neural (:obj:`keras.Layer`): the Keras neural layer.
+        layer_activation (:obj:`keras.Layer`): the Keras activation layer.
+        layer_globalavgpool (:obj:`keras.Layer`): the optional Keras global
             average pooling layer.
 
     Returns:
@@ -204,11 +199,12 @@ def _prepare_threshold_calculations(layer_ak_type,
 
     # Get parameters from the discrete ReLU activation (threshold and step)
     if layer_activation:
-        thresh_keras = (layer_activation.threshold.numpy() *
-                        np.ones(threshold_shape))
-        step_keras = ((layer_activation.step_width.numpy() *
-                       2.**layer_activation.bitwidth / 16) *
-                      np.ones(threshold_shape))
+        thresh_keras = (layer_activation.threshold.numpy() * np.ones(threshold_shape))
+        step_keras = (
+            (layer_activation.step.numpy() *
+             2.**layer_activation.bitwidth /
+             16) *
+            np.ones(threshold_shape))
 
     # Get global average pooling factor
     if isinstance(layer_globalavgpool, layers.GlobalAveragePooling2D):
@@ -231,9 +227,9 @@ def _convert_layer_thresholds(layer_ak_type, layer_neural, layer_activation,
 
     Args:
         layer_ak_type (:obj:`akida.LayerType`): the type of the Akida layer.
-        layer_neural (:obj:`tf.keras.Layer`): the Keras neural layer.
-        layer_activation (:obj:`tf.keras.Layer`): the Keras activation layer.
-        layer_globalavgpool (:obj:`tf.keras.Layer`): the optional Keras global
+        layer_neural (:obj:`keras.Layer`): the Keras neural layer.
+        layer_activation (:obj:`keras.Layer`): the Keras activation layer.
+        layer_globalavgpool (:obj:`keras.Layer`): the optional Keras global
             average pooling layer.
         input_scaling (2-element tuple): the input scaling of the current layer.
         weights_scale_factor (float): the weights scale factor of the current
